function [T,Tchi,Sent] = getTDconv_Ex (Inrg,Sz,beta0,varargin)
% < Description >
%
% [T,Tchi,Sent] = getTDconv_Ex (Inrg,Sz,beta0 [, S_A0L])
%
% Compute thermodynamic properties T*chi (= temperature * static spin
% susceptibility) and entropy by using the conventional NRG method (see
% Sec. III A 1. in Bulla2008 [R. Bulla et al., Rev. Mod. Phys. 80, 395
% (2008)]).
% Here each set of energy levels in the iteration 'n' (so-called "shell")
% is considered as the restricted but effective set of energy eigenstates
% responsible for temperature Inrg.EScale/beta0. Then we can compute
% the expectation value < .. > of operators (e.g., spin-z operator S_z)
% and the partition function Z as the sum of the Boltzmann factors, for the
% spin susceptibility chi is given by
%
%   T*chi = < S_z^2 > - < S_z >^2 
%     (see the equation just before Eq. (47) in Bulla2008)
%
% and the entropy is given by
%
%   S = < H > / T + ln (Z)
%     (see Eq. (50) in Bulla2008)
%
% These properties are for the whole system. So the impurity contribution
% to the spin susceptiblity and the entropy can be obtained by substracting
% the properties of only the bath (by running the NRG for only the bath in
% the absence the impurity) from those of the full system (by runing the
% NRG for the target impurity model).
%
% < Input >
% Inrg : [struct] Struct containing NRG result, as the output from function
%        NRG_IterDiag.
% Sz : [rank-3 tensor] Spin-z operator along the chain.
% beta0 : [number] Prefactor to define the effective tempearture T for all
%       the iterations which is an output as well; see below for detail.
%       The values of temperature associated with the iterations are given
%       by:
%       T = [Inrg.EScale(2)*sqrt(Inrg.Lambda), Inrg.EScale(2:end)]/beta0;
%       Here Inrg.EScale is the energy scale of the iterations which is
%       used to rescale the energy eigenvalues along the iterative
%       diagonalization. And Inrg.Lambda is the logarithmic discretization
%       parameter Lambda. By the convention of NRG_IterDiag, Inrg.EScale(1)
%       is always 1 so that one can easily read off the energy eigenvalues
%       of the impurity without considering prefactor. Therefore, to have a
%       consistent logarithmic scaling along T, the first element T(1) is
%       determined by multiplying sqrt(Inrg.Lambda) to Inrg.EScale(2).
% S_A0L : [rank-3 tensor] (Optional) Spin-z operator acting onto the space
%       associated with the left leg (i.e., first leg) of A0 (the input
%       isometry to NRG_IterDiag).
%       If Simp is not given as input and the left-leg space of A0 is not
%       vacuum (i.e., the associated dimension is larger than 1), the input
%       Sz is regarded as the operator acting onto the left-leg space of
%       A0. If the left-leg space is vacuum (i.e., the associated dimension
%       is 1), the spin operator is not considered for the space.
%
% < Output >
% T : [vector] Temperature values corresponding to shell indices. It is in
%       descending order.
% Tchi : [vector] Temperature * static spin susceptibility. In the Kondo
%       regime, it should decay linearly with T.
% Sent : [vector] Entropy.
%
% Written by S.Lee (May 05,2017); edited by S.Lee (Aug.09,2017)
% Updated by S.Lee (May 12,2019): Revised for SoSe 2019.
% Updated by S.Lee (Jun.20,2020): Revised for SoSe 2020.

% parsing optional input
S_A0L = [];
if ~isempty(varargin)
    S_A0L = varargin{1};
end

% % % sanity check of input
if ndims(Sz) ~= 3
    error('ERR: Sz should be rank-3; use S(:,3,:) for the tensor S generated by getLocalSpace.');
elseif size(Sz,2) ~= 1
    error('ERR: the 2nd dimension of Sz should be singleton.');
elseif size(Sz,1) ~= size(Sz,3)
    error('ERR: the 1st and 3rd dimensions of Sz should be the same.');
elseif isempty(S_A0L) && (size(Inrg.AK{1},1) > 1) && (size(Inrg.AK{1},1) ~= size(Sz,3))
    error('ERR: the dimension of Sz should be consistent with the dimension of the left-leg space of A0.');
end
if ~isempty(S_A0L)
    if ndims(S_A0L) ~= 3
        error('ERR: S_L00 should be rank-3; use S(:,3,:) for the tensor S generated by getLocalSpace.');
    elseif size(S_A0L,2) ~= 1
        error('ERR: the 2nd dimension of S_A0L should be singleton.');
    elseif size(S_A0L,1) ~= size(S_A0L,3)
        error('ERR: the 1st and 3rd dimensions of S_A0L should be the same.');
    elseif size(S_A0L,3) ~= size(Inrg.AK{1},1)
        error('ERR: the dimension of S_A0L should be consistent with the dimension of the left-leg space of A0.');
    end
end
% % % % % 

N = numel(Inrg.EScale); % total number of iterations
T = [Inrg.EScale(2)*sqrt(Inrg.Lambda), Inrg.EScale(2:end)]/beta0; % temperature

% properties before even-odd averaging
Tchi0 = zeros(1,N); % T*(spin susceptibility)
Sent0 = zeros(1,N); % entropy

for itN = (1:N)
    if itN == 1
        % % For the bottom-leg space of A0
        % Sz operator in kept state basis
        SzK = updateLeft([],[],Inrg.AK{itN},Sz,3,Inrg.AK{itN});
        % Sz operator in discarded state basis
        SzD = updateLeft([],[],Inrg.AD{itN},Sz,3,Inrg.AD{itN});
        % NOTE: here we consider all the states in a single iteration; both
        % kept and discarded states. By doing this, we can cover a wider
        % range of energy levels to reduce the numerical artifact
        % originating from finite energy threshold.
        
        % % For the left-leg space of A0:
        % If not vacuum, add the spin operator acting on the space
        if size(Inrg.AK{itN},1) > 1
            if ~isempty(S_A0L)
                SzK = SzK + updateLeft(S_A0L,3,Inrg.AK{itN},[],[],Inrg.AK{itN});
            else
                SzK = SzK + updateLeft(Sz,3,Inrg.AK{itN},[],[],Inrg.AK{itN});
            end
        end
        if size(Inrg.AD{itN},1) > 1
            if ~isempty(S_A0L)
                SzD = SzD + updateLeft(S_A0L,3,Inrg.AD{itN},[],[],Inrg.AD{itN});
            else
                SzD = SzD + updateLeft(Sz,3,Inrg.AD{itN},[],[],Inrg.AD{itN});
            end
        end
    else
        % % % TODO - Exercise (a) (Start) % % %
        % Obtain the Sz operators in kept/discarded state basis. Use
        % SzKprev which is the Sz operator in the kept state basis at the
        % last iteration (see below for its definition), and the Sz
        % operator at the current iteration.
        
        % SzK: Sz operator in kept state basis
        
        % SzD: Sz operator in discarded state basis
        
        % % % TODO - Exercise (a) (End) % % %
    end
    
    % matrix of the Sz operator for the kept+discarded space
    Sztot = blkdiag(permute(SzK,[1 3 2]),permute(SzD,[1 3 2]));
    % third dimensions of SzK and SzD are singleton
    E = [Inrg.EK{itN};Inrg.ED{itN}]; % rescaled energy eigenvalues within a shell
    
    % % % TODO - Exercise (a) (Start) % % %
    % Obtain the temperature * spin susceptibility Tchi0(itN) and the
    % entropy Sent0(itN) for the current iteration
    % NOTE: be aware of that the E above is *rescaled* energy values.
    
    
    % % % TODO - Exercise (a) (End) % % %
    
    SzKprev = SzK;
end

% even-odd averaging
Tchi = (interp1(T(1:2:end),Tchi0(1:2:end),T,'linear','extrap') + ...
    interp1(T(2:2:end),Tchi0(2:2:end),T,'linear','extrap'))/2;
Sent = (interp1(T(1:2:end),Sent0(1:2:end),T,'linear','extrap') + ...
    interp1(T(2:2:end),Sent0(2:2:end),T,'linear','extrap'))/2;

end